@@ -17,15 +17,15 @@
 import tensorflow as tf
 
 
-def vqt(qhbm, num_samples, hamiltonian, beta):
+def vqt(model, num_samples, hamiltonian, beta):
   """Computes the VQT loss of a given QHBM against given thermal state params.
 
   This function is differentiable within a `tf.GradientTape` scope.
 
   Args:
-    qhbm: A `qhbm.QHBM` which is the model whose loss is to be calculated.
+    model: A `qhbm.QHBM` which is the model whose loss is to be calculated.
     num_samples: A scalar `tf.Tensor` specifying the number of samples to draw
-      from the EBM of `qhbm` when estimating the loss and its gradients.
+      from the EBM of `model` when estimating the loss and its gradients.
     hamiltonian: 1D tensor of strings with one entry, the result of calling
       `tfq.convert_to_tensor` on a list containing one cirq.PauliSum, `[op]`.
       Here, `op` is the Hamiltonian against which the loss is calculated.
@@ -37,30 +37,28 @@ def vqt(qhbm, num_samples, hamiltonian, beta):
   """
 
   @tf.custom_gradient
-  def loss(variables):
-    bitstrings, counts = qhbm.ebm.sample(num_samples)
+  def loss(unused):
+    bitstrings, counts = model.ebm.sample(num_samples)
     probs = tf.cast(counts, tf.float32) / tf.cast(num_samples, tf.float32)
     expectation = tf.squeeze(
-        qhbm.qnn.expectation(bitstrings, counts, hamiltonian), -1)
-    if qhbm.is_analytic:
-      entropy = qhbm.entropy()
+        model.qnn.expectation(bitstrings, counts, hamiltonian), -1)
+    if model.is_analytic:
+      entropy = model.entropy()
     else:
       entropy = -tf.reduce_sum(probs * tf.math.log(probs))
 
     def grad(grad_y, variables=None):
-      with tf.GradientTape() as qnn_tape:
+      with tf.GradientTape() as tape:
         beta_expectations = beta * tf.squeeze(
-            qhbm.qnn.expectation(bitstrings, counts, hamiltonian, reduce=False),
-            -1)
+            model.qnn.expectation(
+                bitstrings, counts, hamiltonian, reduce=False), -1)
         beta_expectation = tf.reduce_sum(probs * beta_expectations)
-      grad_qnn = qnn_tape.gradient(beta_expectation,
-                                   qhbm.qnn.trainable_variables)
+      grad_qnn = tape.gradient(beta_expectation, model.phis)
       grad_qnn = [grad_y * grad for grad in grad_qnn]
 
-      with tf.GradientTape() as ebm_tape:
-        energies = qhbm.ebm.energy(bitstrings)
-      energy_gradients = ebm_tape.jacobian(energies,
-                                           qhbm.ebm.trainable_variables)
+      with tf.GradientTape() as tape:
+        energies = model.ebm.energy(bitstrings)
+      energy_gradients = tape.jacobian(energies, model.thetas)
       probs_diffs = probs * (beta_expectations - energies)
       avg_diff = tf.reduce_sum(probs_diffs)
       grad_ebm = [
@@ -71,8 +69,10 @@ def vqt(qhbm, num_samples, hamiltonian, beta):
           for grad in energy_gradients
       ]
       grad_vars = grad_ebm + grad_qnn
-      return grad_vars
+      if variables is None:
+        return grad_vars
+      return grad_vars, [tf.zeros_like(g) for g in grad_vars]
 
     return beta * expectation - entropy, grad
 
-  return loss(qhbm.trainable_variables)
+  return loss(model.trainable_variables)
